@@ -52,7 +52,7 @@ use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
 BEGIN {
     # Declare @ISA, $VERSION, $GLOBAL_CONTEXT_ARGS
     @ISA = qw(IO::Socket::INET);
-    $VERSION = '1.11';
+    $VERSION = '1.13';
     $GLOBAL_CONTEXT_ARGS = {};
 
     #Make $DEBUG another name for $Net::SSLeay::trace
@@ -68,7 +68,14 @@ BEGIN {
 
 }
 
-sub CLONE_SKIP { 1 }
+sub DEBUG {
+    $DEBUG or return;
+    my (undef,$file,$line) = caller;
+    my $msg = shift;
+    $msg = sprintf $msg,@_ if @_;
+    print STDERR "DEBUG: $file:$line: $msg\n";
+}
+
 
 # Export some stuff
 # inet4|inet6|debug will be handeled by myself, everything
@@ -235,7 +242,8 @@ sub connect_SSL {
     my $timeout = exists $args->{Timeout} 
     	? $args->{Timeout} 
 	: ${*$self}{io_socket_timeout}; # from IO::Socket
-    if ( defined($timeout) && $timeout>=0 && $self->blocking(0) ) {
+    if ( defined($timeout) && $timeout>0 && $self->blocking(0) ) {
+	#DEBUG( "set socket to non-blocking to enforce timeout=$timeout" );
 	# timeout was given and socket was blocking
     	# enforce timeout with now non-blocking socket
     } else {
@@ -247,12 +255,13 @@ sub connect_SSL {
     for my $dummy (1) {
 	#DEBUG( 'calling ssleay::connect' );
 	my $rv = Net::SSLeay::connect($ssl);
-	#DEBUG( "rv=$rv" );
+	#DEBUG( "connect -> rv=$rv" );
 	if ( $rv < 0 ) {
 	    unless ( $self->_set_rw_error( $ssl,$rv )) {
 		$self->error("SSL connect attempt failed with unknown error");
 		delete ${*$self}{'_SSL_opening'};
 		${*$self}{'_SSL_opened'} = 1;
+		#DEBUG( "fatal SSL error: $SSL_ERROR" );
 		return $self->fatal_ssl_error();
 	    }
 
@@ -266,14 +275,17 @@ sub connect_SSL {
 	    if ( $timeout>0 ) {
 		my $vec = '';
 		vec($vec,$self->fileno,1) = 1;
+		#DEBUG( "waiting for fd to become ready: $SSL_ERROR" );
 	    	$rv = 
 		    $SSL_ERROR == SSL_WANT_READ ? select( $vec,undef,undef,$timeout) :
 		    $SSL_ERROR == SSL_WANT_WRITE ? select( undef,$vec,undef,$timeout) :
 		    undef;
 	    } else {
+		#DEBUG( "handshake failed because no more time" );
 	    	$! = ETIMEDOUT
 	    }
 	    if ( ! $rv ) {
+		#DEBUG( "handshake failed because socket did not became ready" );
 		# failed because of timeout, return
 	    	$! ||= ETIMEDOUT;
 		delete ${*$self}{'_SSL_opening'};
@@ -283,6 +295,7 @@ sub connect_SSL {
 	    }
 
 	    # socket is ready, try non-blocking connect again after recomputing timeout
+	    #DEBUG( "socket ready, retrying connect" );
 	    my $now = time();
 	    $timeout -= $now - $start;
 	    $start = $now;
@@ -290,6 +303,7 @@ sub connect_SSL {
 
 	} elsif ( $rv == 0 ) {
 	    delete ${*$self}{'_SSL_opening'};
+	    #DEBUG( "connection failed - connect returned 0" );
 	    $self->error("SSL connect attempt failed because of handshake problems" );
 	    ${*$self}{'_SSL_opened'} = 1;
 	    return $self->fatal_ssl_error();
@@ -385,7 +399,7 @@ sub accept_SSL {
     my $timeout = exists $args->{Timeout} 
     	? $args->{Timeout} 
 	: ${*$self}{io_socket_timeout}; # from IO::Socket
-    if ( defined($timeout) && $timeout>=0 && $socket->blocking(0) ) {
+    if ( defined($timeout) && $timeout>0 && $socket->blocking(0) ) {
 	# timeout was given and socket was blocking
     	# enforce timeout with now non-blocking socket
     } else {
@@ -853,6 +867,7 @@ sub errstr {
 sub fatal_ssl_error {
     my $self = shift;
     my $error_trap = ${*$self}{'_SSL_arguments'}->{'SSL_error_trap'};
+    $@ = $self->errstr;
     if (defined $error_trap and ref($error_trap) eq 'CODE') {
 	$error_trap->($self, $self->errstr()."\n".$self->get_ssleay_error());
     } else { $self->kill_socket; }
@@ -949,7 +964,6 @@ BEGIN {
     $HAVE_WEAKREF = $@ ? 0 : 1;
 }
 
-sub CLONE_SKIP { 1 }
 
 sub TIEHANDLE {
     my ($class, $handle) = @_;
@@ -980,17 +994,20 @@ sub CLOSE {                          #<---- Do not change this function!
 package IO::Socket::SSL::SSL_Context;
 use strict;
 
+my %CTX_CREATED_IN_THIS_THREAD;
+*DEBUG = *IO::Socket::SSL::DEBUG;
+
 # should be better taken from Net::SSLeay, but they are not (yet) defined there
 use constant SSL_MODE_ENABLE_PARTIAL_WRITE => 1;
 use constant SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER => 2;
 
-sub CLONE_SKIP { 1 }
 
 # Note that the final object will actually be a reference to the scalar
 # (C-style pointer) returned by Net::SSLeay::CTX_*_new() so that
 # it can be blessed.
 sub new {
     my $class = shift;
+    DEBUG( "$class @_" );
     my $arg_hash = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
 
     my $ctx_object = $arg_hash->{'SSL_reuse_ctx'};
@@ -1108,6 +1125,9 @@ sub new {
     Net::SSLeay::CTX_set_verify($ctx, $verify_mode, $verify_callback);
 
     $ctx_object = { context => $ctx };
+    DEBUG( "new ctx $ctx" );
+    $CTX_CREATED_IN_THIS_THREAD{$ctx} = 1;
+
     if ( my $cache = $arg_hash->{SSL_session_cache} ) {
 	# use predefined cache
     	$ctx_object->{session_cache} = $cache
@@ -1136,18 +1156,23 @@ sub has_session_cache {
 }
 
 
+sub CLONE { %CTX_CREATED_IN_THIS_THREAD = (); DEBUG( "clone!" ) }
 sub DESTROY {
     my $self = shift;
-    $self->{context} and Net::SSLeay::CTX_free($self->{context});
+    if ( my $ctx = $self->{context} ) {
+	DEBUG( "free ctx $ctx open=".join( " ",keys %CTX_CREATED_IN_THIS_THREAD ));
+	if ( %CTX_CREATED_IN_THIS_THREAD and 
+	    delete $CTX_CREATED_IN_THIS_THREAD{$ctx} ) {
+	    DEBUG( "OK free ctx $ctx" );
+	    Net::SSLeay::CTX_free($ctx);
+	}
+    }
     delete(@{$self}{'context','session_cache'});
 }
 
-
 package IO::Socket::SSL::Session_Cache;
 use strict;
 
-sub CLONE_SKIP { 1 }
-
 sub new {
     my ($class, $size) = @_;
     $size>0 or return;
